#ifndef SBCLIB_GFX_OPENGL
// The OS-specific part of the interface with OpenGL
// Windows part based on tutoral http://www.essi.fr/~buffa/cours/synthese_image/DOCS/Tutoriaux/Nehe/lesson%d.htm
// Includes making the window (Linux part uses SDL), setting the mode, res etc.
// Link OpenGL32.lib GLu32.lib for Windows
// Link -lSomeOtherStuff for Linux
#ifdef __APPLE__
#define linux
#endif
#ifndef linux
#include <windows.h>	// Header File For Windows
#endif
#include <rnd/macros.c> // Define WINMAIN, DEBUG etc.; KEY, LOOP are in this file
#include <GL/gl.h>		// Header File For The OpenGL32 Library
#include <GL/glu.h>		// Header File For The GLu32 Library

// Globals interfacing with the higher-level stuff
int xres,yres;
unsigned char keystate[384],oldkeystate[384];		// Array Used For The Keyboard Routine
char gotch[500]="";
#ifdef SBCLIB_GRAPHICS_WINDOWED
char gl_windowed=1;
#else
char gl_windowed=0;
#endif

void ReSizeGLScene(const int xs,int ys); // Generic functions needed from <graphics_opengl.c>
void InitGL(const int xs,const int ys);

#include <rnd/sjbdatetime.c>
void gwarn(const char *message)
{ // Called by gerror, which depends on the OS
	FILE *out=fopen("graphics_error.log","at");
	fprintf(out,"%s %s\n",sjbdatetime(),message);
	fclose(out);
}

// Macros identical between Linux and Windows
#define KEYUP(x) while (!dealwithmessages() && KEY(x));

#define LOOP if (!xres) gerror("GRAPHICS() macro not called at LOOP start, or no resolution specified"); \
	char leave; for (leave=(dealwithmessages(),0);!leave;leave|=dealwithmessages()|KEY(VK_Q)|KEY(VK_ESCAPE))

#define LOOPESC char leave; for (leave=(dealwithmessages(),0);!leave;leave|=dealwithmessages()|KEY(VK_ESCAPE))
#define PAUSE for (dealwithmessages();!dealwithmessages() && !KHIT(VK_SPACE););

#ifdef linux
// Now trying this using SDL
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#ifdef SBCLIB_GRAPHICS_WINDOWED
#define SBCLIB_SDLvidmodeflags SDL_OPENGL
#else
#define SBCLIB_SDLvidmodeflags SDL_OPENGL|SDL_FULLSCREEN
#endif

#define GRAPHICS(X,Y,T) \
{ \
	int xs=X,ys=Y; \
	if (SDL_Init(SDL_INIT_EVERYTHING)<0) return; \
	if (xs==0) \
	{ \
		const SDL_VideoInfo *vi=SDL_GetVideoInfo(); \
		xs=vi->current_w; ys=vi->current_h; \
	} \
	if (!SDL_SetVideoMode(xs,ys,32,SBCLIB_SDLvidmodeflags)) return; \
	InitGL(xs,ys); \
	SDL_WM_SetCaption(T,NULL); \
	memset(oldkeystate,0,384); memset(keystate,0,384); \
}

#define wglGetProcAddress glXGetProcAddress

#define KEY(x) (keystate[x])
#define KHIT(x) (keystate[x]&~oldkeystate[x])

void gracefulexit(void) {SDL_Quit(); exit(0);}

void gerror(const char *message)
{
	gwarn(message);
	printf("\n%s\n",message);
	gracefulexit();
}

int dealwithmessages(void)
{ // Errr, looks like the Linux/SDL version doesn't poll the mouse position or buttons yet
	SDL_Event event;
	memcpy(oldkeystate,keystate,384);
	while (SDL_PollEvent(&event)) switch (event.type)
	{
		case SDL_KEYDOWN: keystate[event.key.keysym.sym]=1; break;
		case SDL_KEYUP: keystate[event.key.keysym.sym]=0; break;
		case SDL_QUIT: gracefulexit(); break;
	}
	return 0;
}

// Program for generating the following stuff
/*// gcc vmemtest.c -I ~/include -o vmemtest
#include <stdio.h>
#include <rnd/fgetline.c>
#include <rnd/streqi.c>

int main(void)
{
	FILE *in=fopen("/usr/include/SDL/SDL_keysym.h","rb");
	char *l,*s,*t;
	while (l=fgetline(in))
	{
		if (s=strstr(l,"SDLK_"))
		{
			if (t=strchr(s,'\t'))
			{
				*t=0;
				char *u; streqi(&u,s);
				for (t=u;*t;t++) if (*t>='a' && *t<='z') *t-=32;
				printf("#define VK_%s SDLK_%s\n",u+5,s+5);
				free(u);
			}
		}
		free(l);
	}
	return 0;
}*/

// Code autogenerated by replacements in the SDL_keyboard.h header
#define VK_BACKSPACE SDLK_BACKSPACE
#define VK_TAB SDLK_TAB
#define VK_CLEAR SDLK_CLEAR
#define VK_RETURN SDLK_RETURN
#define VK_PAUSE SDLK_PAUSE
#define VK_ESCAPE SDLK_ESCAPE
#define VK_SPACE SDLK_SPACE
#define VK_EXCLAIM SDLK_EXCLAIM
#define VK_QUOTEDBL SDLK_QUOTEDBL
#define VK_HASH SDLK_HASH
#define VK_DOLLAR SDLK_DOLLAR
#define VK_AMPERSAND SDLK_AMPERSAND
#define VK_QUOTE SDLK_QUOTE
#define VK_LEFTPAREN SDLK_LEFTPAREN
#define VK_RIGHTPAREN SDLK_RIGHTPAREN
#define VK_ASTERISK SDLK_ASTERISK
#define VK_PLUS SDLK_PLUS
#define VK_COMMA SDLK_COMMA
#define VK_MINUS SDLK_MINUS
#define VK_PERIOD SDLK_PERIOD
#define VK_SLASH SDLK_SLASH
#define VK_0 SDLK_0
#define VK_1 SDLK_1
#define VK_2 SDLK_2
#define VK_3 SDLK_3
#define VK_4 SDLK_4
#define VK_5 SDLK_5
#define VK_6 SDLK_6
#define VK_7 SDLK_7
#define VK_8 SDLK_8
#define VK_9 SDLK_9
#define VK_COLON SDLK_COLON
#define VK_SEMICOLON SDLK_SEMICOLON
#define VK_LESS SDLK_LESS
#define VK_EQUALS SDLK_EQUALS
#define VK_GREATER SDLK_GREATER
#define VK_QUESTION SDLK_QUESTION
#define VK_AT SDLK_AT
#define VK_LEFTBRACKET SDLK_LEFTBRACKET
#define VK_BACKSLASH SDLK_BACKSLASH
#define VK_RIGHTBRACKET SDLK_RIGHTBRACKET
#define VK_CARET SDLK_CARET
#define VK_UNDERSCORE SDLK_UNDERSCORE
#define VK_BACKQUOTE SDLK_BACKQUOTE
#define VK_A SDLK_a
#define VK_B SDLK_b
#define VK_C SDLK_c
#define VK_D SDLK_d
#define VK_E SDLK_e
#define VK_F SDLK_f
#define VK_G SDLK_g
#define VK_H SDLK_h
#define VK_I SDLK_i
#define VK_J SDLK_j
#define VK_K SDLK_k
#define VK_L SDLK_l
#define VK_M SDLK_m
#define VK_N SDLK_n
#define VK_O SDLK_o
#define VK_P SDLK_p
#define VK_Q SDLK_q
#define VK_R SDLK_r
#define VK_S SDLK_s
#define VK_T SDLK_t
#define VK_U SDLK_u
#define VK_V SDLK_v
#define VK_W SDLK_w
#define VK_X SDLK_x
#define VK_Y SDLK_y
#define VK_Z SDLK_z
#define VK_DELETE SDLK_DELETE
#define VK_WORLD_0 SDLK_WORLD_0
#define VK_WORLD_1 SDLK_WORLD_1
#define VK_WORLD_2 SDLK_WORLD_2
#define VK_WORLD_3 SDLK_WORLD_3
#define VK_WORLD_4 SDLK_WORLD_4
#define VK_WORLD_5 SDLK_WORLD_5
#define VK_WORLD_6 SDLK_WORLD_6
#define VK_WORLD_7 SDLK_WORLD_7
#define VK_WORLD_8 SDLK_WORLD_8
#define VK_WORLD_9 SDLK_WORLD_9
#define VK_WORLD_10 SDLK_WORLD_10
#define VK_WORLD_11 SDLK_WORLD_11
#define VK_WORLD_12 SDLK_WORLD_12
#define VK_WORLD_13 SDLK_WORLD_13
#define VK_WORLD_14 SDLK_WORLD_14
#define VK_WORLD_15 SDLK_WORLD_15
#define VK_WORLD_16 SDLK_WORLD_16
#define VK_WORLD_17 SDLK_WORLD_17
#define VK_WORLD_18 SDLK_WORLD_18
#define VK_WORLD_19 SDLK_WORLD_19
#define VK_WORLD_20 SDLK_WORLD_20
#define VK_WORLD_21 SDLK_WORLD_21
#define VK_WORLD_22 SDLK_WORLD_22
#define VK_WORLD_23 SDLK_WORLD_23
#define VK_WORLD_24 SDLK_WORLD_24
#define VK_WORLD_25 SDLK_WORLD_25
#define VK_WORLD_26 SDLK_WORLD_26
#define VK_WORLD_27 SDLK_WORLD_27
#define VK_WORLD_28 SDLK_WORLD_28
#define VK_WORLD_29 SDLK_WORLD_29
#define VK_WORLD_30 SDLK_WORLD_30
#define VK_WORLD_31 SDLK_WORLD_31
#define VK_WORLD_32 SDLK_WORLD_32
#define VK_WORLD_33 SDLK_WORLD_33
#define VK_WORLD_34 SDLK_WORLD_34
#define VK_WORLD_35 SDLK_WORLD_35
#define VK_WORLD_36 SDLK_WORLD_36
#define VK_WORLD_37 SDLK_WORLD_37
#define VK_WORLD_38 SDLK_WORLD_38
#define VK_WORLD_39 SDLK_WORLD_39
#define VK_WORLD_40 SDLK_WORLD_40
#define VK_WORLD_41 SDLK_WORLD_41
#define VK_WORLD_42 SDLK_WORLD_42
#define VK_WORLD_43 SDLK_WORLD_43
#define VK_WORLD_44 SDLK_WORLD_44
#define VK_WORLD_45 SDLK_WORLD_45
#define VK_WORLD_46 SDLK_WORLD_46
#define VK_WORLD_47 SDLK_WORLD_47
#define VK_WORLD_48 SDLK_WORLD_48
#define VK_WORLD_49 SDLK_WORLD_49
#define VK_WORLD_50 SDLK_WORLD_50
#define VK_WORLD_51 SDLK_WORLD_51
#define VK_WORLD_52 SDLK_WORLD_52
#define VK_WORLD_53 SDLK_WORLD_53
#define VK_WORLD_54 SDLK_WORLD_54
#define VK_WORLD_55 SDLK_WORLD_55
#define VK_WORLD_56 SDLK_WORLD_56
#define VK_WORLD_57 SDLK_WORLD_57
#define VK_WORLD_58 SDLK_WORLD_58
#define VK_WORLD_59 SDLK_WORLD_59
#define VK_WORLD_60 SDLK_WORLD_60
#define VK_WORLD_61 SDLK_WORLD_61
#define VK_WORLD_62 SDLK_WORLD_62
#define VK_WORLD_63 SDLK_WORLD_63
#define VK_WORLD_64 SDLK_WORLD_64
#define VK_WORLD_65 SDLK_WORLD_65
#define VK_WORLD_66 SDLK_WORLD_66
#define VK_WORLD_67 SDLK_WORLD_67
#define VK_WORLD_68 SDLK_WORLD_68
#define VK_WORLD_69 SDLK_WORLD_69
#define VK_WORLD_70 SDLK_WORLD_70
#define VK_WORLD_71 SDLK_WORLD_71
#define VK_WORLD_72 SDLK_WORLD_72
#define VK_WORLD_73 SDLK_WORLD_73
#define VK_WORLD_74 SDLK_WORLD_74
#define VK_WORLD_75 SDLK_WORLD_75
#define VK_WORLD_76 SDLK_WORLD_76
#define VK_WORLD_77 SDLK_WORLD_77
#define VK_WORLD_78 SDLK_WORLD_78
#define VK_WORLD_79 SDLK_WORLD_79
#define VK_WORLD_80 SDLK_WORLD_80
#define VK_WORLD_81 SDLK_WORLD_81
#define VK_WORLD_82 SDLK_WORLD_82
#define VK_WORLD_83 SDLK_WORLD_83
#define VK_WORLD_84 SDLK_WORLD_84
#define VK_WORLD_85 SDLK_WORLD_85
#define VK_WORLD_86 SDLK_WORLD_86
#define VK_WORLD_87 SDLK_WORLD_87
#define VK_WORLD_88 SDLK_WORLD_88
#define VK_WORLD_89 SDLK_WORLD_89
#define VK_WORLD_90 SDLK_WORLD_90
#define VK_WORLD_91 SDLK_WORLD_91
#define VK_WORLD_92 SDLK_WORLD_92
#define VK_WORLD_93 SDLK_WORLD_93
#define VK_WORLD_94 SDLK_WORLD_94
#define VK_WORLD_95 SDLK_WORLD_95
#define VK_KP0 SDLK_KP0
#define VK_KP1 SDLK_KP1
#define VK_KP2 SDLK_KP2
#define VK_KP3 SDLK_KP3
#define VK_KP4 SDLK_KP4
#define VK_KP5 SDLK_KP5
#define VK_KP6 SDLK_KP6
#define VK_KP7 SDLK_KP7
#define VK_KP8 SDLK_KP8
#define VK_KP9 SDLK_KP9
#define VK_KP_PERIOD SDLK_KP_PERIOD
#define VK_KP_DIVIDE SDLK_KP_DIVIDE
#define VK_KP_MULTIPLY SDLK_KP_MULTIPLY
#define VK_KP_MINUS SDLK_KP_MINUS
#define VK_KP_PLUS SDLK_KP_PLUS
#define VK_KP_ENTER SDLK_KP_ENTER
#define VK_KP_EQUALS SDLK_KP_EQUALS
#define VK_UP SDLK_UP
#define VK_DOWN SDLK_DOWN
#define VK_RIGHT SDLK_RIGHT
#define VK_LEFT SDLK_LEFT
#define VK_INSERT SDLK_INSERT
#define VK_HOME SDLK_HOME
#define VK_END SDLK_END
#define VK_PAGEUP SDLK_PAGEUP
#define VK_PAGEDOWN SDLK_PAGEDOWN
#define VK_F1 SDLK_F1
#define VK_F2 SDLK_F2
#define VK_F3 SDLK_F3
#define VK_F4 SDLK_F4
#define VK_F5 SDLK_F5
#define VK_F6 SDLK_F6
#define VK_F7 SDLK_F7
#define VK_F8 SDLK_F8
#define VK_F9 SDLK_F9
#define VK_F10 SDLK_F10
#define VK_F11 SDLK_F11
#define VK_F12 SDLK_F12
#define VK_F13 SDLK_F13
#define VK_F14 SDLK_F14
#define VK_F15 SDLK_F15
#define VK_NUMLOCK SDLK_NUMLOCK
#define VK_CAPSLOCK SDLK_CAPSLOCK
#define VK_SCROLLOCK SDLK_SCROLLOCK
#define VK_RSHIFT SDLK_RSHIFT
#define VK_LSHIFT SDLK_LSHIFT
#define VK_RCTRL SDLK_RCTRL
#define VK_LCTRL SDLK_LCTRL
#define VK_RALT SDLK_RALT
#define VK_LALT SDLK_LALT
#define VK_RMETA SDLK_RMETA
#define VK_LMETA SDLK_LMETA
#define VK_LSUPER SDLK_LSUPER
#define VK_RSUPER SDLK_RSUPER
#define VK_MODE SDLK_MODE
#define VK_COMPOSE SDLK_COMPOSE
#define VK_HELP SDLK_HELP
#define VK_PRINT SDLK_PRINT
#define VK_SYSREQ SDLK_SYSREQ
#define VK_BREAK SDLK_BREAK
#define VK_MENU SDLK_MENU
#define VK_POWER SDLK_POWER
#define VK_EURO SDLK_EURO
#define VK_UNDO SDLK_UNDO
// END autogenerated code

#define VK_ADD VK_KP_PLUS
#define VK_SUBTRACT VK_KP_MINUS

#else // Not linux therefore Windows

static HGLRC hRC;		// Permanent Rendering Context
static HDC hDC;		// Private GDI Device Context
HINSTANCE hinstance; HWND winhandle;

int mx=0,my=0,mb_real,mb_stuck=1,mb=0;

#define GRAPHICS(X,Y,T) if (!makethewindow(X,Y,T)) return;

#define KEY(x) (keystate[x]>>7)
#define KHIT(x) ((keystate[x]&~oldkeystate[x])>>7)

#define VK_ALT VK_MENU
#define VK_LALT VK_LMENU
#define VK_RALT VK_RMENU

#ifndef __LCC__
#define VK_0 '0'
#define VK_1 '1'
#define VK_2 '2'
#define VK_3 '3'
#define VK_4 '4'
#define VK_5 '5'
#define VK_6 '6'
#define VK_7 '7'
#define VK_8 '8'
#define VK_9 '9'
#endif

void gracefulexit(void)
{
	if (!gl_windowed) ChangeDisplaySettings(NULL,0);
	wglMakeCurrent(hDC,NULL);
	wglDeleteContext(hRC);
	ReleaseDC(winhandle,hDC);
	PostQuitMessage(0);
	exit(0);
}

void gerror(const char *message)
{
	gwarn(message);
	MessageBox(NULL,message,"Oh sod it",MB_ICONERROR|MB_SETFOREGROUND|MB_OK);
	gracefulexit();
}
//#include <rnd/eithertime.c>
int dealwithmessages(void)
{
/*FILE *out=fopen("dealwithmessages.log","at");
fprintf(out,"[%lf",eithertime());
fclose(out);*/
	mb&=~(4|8); // Reset wheel events
	MSG msg; // Process all messages
	while (PeekMessage(&msg,NULL,0,0,PM_NOREMOVE))
	{
/*out=fopen("dealwithmessages.log","at");
fprintf(out,"(%d){%lf|",msg.message,eithertime());
fclose(out);*/
		if (GetMessage(&msg,NULL,0,0))
		{
/*out=fopen("dealwithmessages.log","at");
fprintf(out,"%lf||",eithertime());
fclose(out);*/
			TranslateMessage(&msg); // 161=WM_NCLBUTTONDOWN, non-client area left button down
			DispatchMessage(&msg); // <-- Multiple seconds disappear in one of these two while dragging, on msg.message=161
/*out=fopen("dealwithmessages.log","at");
fprintf(out,"%lf}",eithertime());
fclose(out);*/
		}
		else return 1;
	}
/*out=fopen("dealwithmessages.log","at");
fprintf(out,",%lf]\n",eithertime());
fclose(out);*/
	static char closing=0;
	if (!IsIconic(winhandle))
	{
		if (!closing && !gl_windowed) SetWindowPos(winhandle,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE); // Always on top (stops Miranda IM appearing on top for instance)
		memcpy(oldkeystate,keystate,384);
		GetKeyboardState(keystate);
		if (KEY(VK_ALT) && KEY(VK_TAB)) closing=1;
		if (closing) CloseWindow(winhandle);
	}
	else closing=0;
	return 0;
}

/*unsigned DefWindowProc_t(MSG *m)
{
	DefWindowProc(m->hwnd,m->message,m->wParam,m->lParam);
	free(m);
	return 0;
}*/

LRESULT CALLBACK windowproc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	switch (message)
	{
		case WM_DESTROY: case WM_CLOSE: // Trigger using SendMessage(winhandle,WM_CLOSE,0,0);
			gracefulexit();
			break;
		//case WM_KEYDOWN: keystate[wParam]=TRUE; break;
		//case WM_KEYUP: keystate[wParam]=FALSE; break; // Commented out because I've implemented this in dealwithmessages()
		case WM_SYSKEYDOWN: return 0; // Stop ALT or F10 activating the window "menu"
		case WM_CHAR:
		{ // Buffering for ordered text input
			//keyhit=1; // Came from <gfx_dx.c>
			int l=strlen(gotch);
			if (l<499)
			{
				gotch[l+1]=0;
				gotch[l]=(char)wParam;
			}
			break;
		}
		case WM_SIZE:
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam)); // Has to be called in thread that actually has OpenGL in it
			break;
		/*case WM_SIZING:
		{
			RECT *pr=(RECT *)lParam;
			ReSizeGLScene(pr->right-pr->left,pr->bottom-pr->top);
		} break;*/
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN: case WM_LBUTTONUP: //case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDOWN: case WM_RBUTTONUP: //case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDOWN: case WM_MBUTTONUP: //case WM_MBUTTONDBLCLK:
			mx=(int)LOWORD(lParam);
			my=(int)HIWORD(lParam);
			mb_real=(int)wParam&(MK_LBUTTON|MK_RBUTTON|MK_MBUTTON);
			if (!mb_real) mb_stuck=0;
			if (!mb_stuck) mb=mb_real;
			break;
		case WM_MOUSEWHEEL:
			if ((int)wParam>0) mb|=4; else mb|=8;
			break;
		/*case WM_NCLBUTTONDOWN:
		{ // Call DefWindowProc in its own thread when it does a blocking (modal) operation, e.g. window resizing
			MSG *m=malloc(sizeof(MSG)); // This doesn't exactly work, still manages to block somehow
			m->hwnd=hWnd; m->message=message; // I could try putting the WHOLE of dealwithmessages in a thread?
			m->wParam=wParam; m->lParam=lParam;
			unsigned long mallards;
			CloseHandle(CreateThread(NULL,0,DefWindowProc_t,m,0,&mallards));
			return 0;
		}*/
		default:
			return (DefWindowProc(hWnd,message,wParam,lParam));
	}
	return 0;
}

HICON makethewindow_icon=NULL;
int makethewindow_shiftx=0,makethewindow_shifty=0; // I'm *guessing* the primary display starts at (0,0)

int makethewindow(const int xs,const int ys,const char *taskname)
{ // Returns 1 on success, 0 on failure
	int wx,wy,wxs,wys;
	if (xs==0)
	{ // Autodetect - primary display
		wx=makethewindow_shiftx; wy=makethewindow_shifty;
		xres=GetSystemMetrics(SM_CXSCREEN);
		yres=GetSystemMetrics(SM_CYSCREEN);
		if (gl_windowed) {xres=(int)(xres*0.6); yres=(int)(yres*0.6);}
		wxs=xres; wys=yres;
	}
	else if (xs==-1)
	{ // Multi-monitor spanning mode
		wx=GetSystemMetrics(SM_XVIRTUALSCREEN);
		wy=GetSystemMetrics(SM_YVIRTUALSCREEN);
		wxs=xres=GetSystemMetrics(SM_CXVIRTUALSCREEN);
		wys=yres=GetSystemMetrics(SM_CYVIRTUALSCREEN);
	}
	else
	{
		wxs=xres=xs; wys=yres=ys;
		wx=wy=0;
	}
	if (gl_windowed)
	{ // Try and make it centred on the mouse pointer but then clipped to the span of the screens
		wxs+=2*GetSystemMetrics(SM_CXFRAME); wys+=GetSystemMetrics(SM_CYCAPTION)+2*GetSystemMetrics(SM_CYFRAME);
		POINT mp; GetCursorPos(&mp);
		wx=max(GetSystemMetrics(SM_XVIRTUALSCREEN),min(mp.x-wxs/2,GetSystemMetrics(SM_XVIRTUALSCREEN)+GetSystemMetrics(SM_CXVIRTUALSCREEN)-wxs));
		wy=max(GetSystemMetrics(SM_YVIRTUALSCREEN),min(mp.y-wys/2,GetSystemMetrics(SM_YVIRTUALSCREEN)+GetSystemMetrics(SM_CYVIRTUALSCREEN)-wys));
	}
	if (xres==0 || yres==0) return 0;
	memset(keystate,0,384); memset(oldkeystate,0,384);
	WNDCLASS wc; // Windows Class Structure Used To Set Up The Type Of Window
	wc.style=CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc=(WNDPROC)windowproc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hinstance;
	wc.hIcon=makethewindow_icon;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName="OpenGL WinClass";
	if (!RegisterClass(&wc)) gerror("Failed To Register The Window Class.");
	winhandle=CreateWindow("OpenGL WinClass",taskname,
		WS_VISIBLE|WS_CLIPCHILDREN|WS_CLIPSIBLINGS|
		(gl_windowed?WS_TILED|WS_SYSMENU|WS_THICKFRAME|WS_MINIMIZEBOX|WS_MAXIMIZEBOX:WS_POPUP),
		wx,wy,wxs,wys,NULL,NULL,hinstance,NULL);
	if (!winhandle) gerror("Window Creation Error.");
	const static PIXELFORMATDESCRIPTOR pfd=
	{
		sizeof(PIXELFORMATDESCRIPTOR),		// Size Of This Pixel Format Descriptor
		1,									// Version Number (?)
		PFD_DRAW_TO_WINDOW				// Format Must Support Window
		| PFD_SUPPORT_OPENGL				// Format Must Support OpenGL
		#ifndef SBCLIB_GRAPHICS_IMMEDIATE
		| PFD_DOUBLEBUFFER					// Must Support Double Buffering
		#endif
		,PFD_TYPE_RGBA,						// Request An RGBA Format
		24,									// Select A 16M Color Depth (changed from 16 in tutorial)
		0, 0, 0, 0, 0, 0,					// Color Bits Ignored (?)
		0,									// No Alpha Buffer
		0,									// Shift Bit Ignored (?)
		0,									// No Accumulation Buffer
		0, 0, 0, 0,							// Accumulation Bits Ignored (?)
		32,									// Highest bitness Z-Buffer available (Depth Buffer)
		0,									// No Stencil Buffer
		0,									// No Auxiliary Buffer (?)
		PFD_MAIN_PLANE,						// Main Drawing Layer
		0,									// Reserved (?)
		0, 0, 0								// Layer Masks Ignored (?)
	};
	hDC=GetDC(winhandle);				// Gets A Device Context For The Window
	int PixelFormat=ChoosePixelFormat(hDC,&pfd);		// Finds The Closest Match To The Pixel Format We Set Above
	if (!PixelFormat) gerror("Can't Find A Suitable PixelFormat.");
	if (!SetPixelFormat(hDC,PixelFormat,&pfd)) gerror("Can't Set The PixelFormat.");
	hRC=wglCreateContext(hDC);
	if (!hRC) gerror("Can't Create A GL Rendering Context.");
	if (!wglMakeCurrent(hDC,hRC)) gerror("Can't activate GLRC.");
	InitGL(xres,yres);
	if (xs>0 && !gl_windowed)
	{
		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings,0,sizeof(DEVMODE));
		dmScreenSettings.dmSize=sizeof(DEVMODE);
		dmScreenSettings.dmPelsWidth=xres; // Screen Width
		dmScreenSettings.dmPelsHeight=yres; // Screen Height
		dmScreenSettings.dmFields=DM_PELSWIDTH|DM_PELSHEIGHT; // Pixel Mode
		ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN); // Switch To Full Screen
	}
	if (!gl_windowed) ShowCursor(0);
	return 1; // Success
}
#endif

#define SBCLIB_GFX_OPENGL
#endif
