#ifndef INCLUDED_SBCLIB_GRAPHICS_OPENGL
// OpenGL needs a different "graphics layer 2" because writing bytes to the framebuffer is no longer fastest
#include <math.h>
#include <gfx_opengl.c> // Link opengl32.lib and glu32.lib for Windows; some other things for Linux
#include <GL/glext.h> // Needed for cubemaps, download from http://www.opengl.org/registry/api/glext.h

char popchar(void)
{
	char ret=*gotch;
	if (ret) for (int n=0;gotch[n];n++) gotch[n]=gotch[n+1];
	return ret;
}

char gl_ortho=1;
#define GRAPHICS3D(X,Y,T) gl_ortho=0; GRAPHICS(X,Y,T)
#define glIsSupported(X) (!!strstr((char *)glGetString(GL_EXTENSIONS),X)) // Warning: this is fairly slow (and not substring-proof)

#define GFX_white 0xFFFFFF
#define RGB666(R,G,B) ((R)*0x330000+(G)*0x3300+(B)*0x33)

void setfovzz(const GLfloat degs,const GLfloat zn,const GLfloat zf)
{ // Write my own that does it in X
	glMatrixMode(GL_PROJECTION);
	GLfloat m[16]; // GL stores matrices column-first
	for (int i=0;i<16;i++) m[i]=0;
	GLfloat xtan=tan(degs degrees*0.5),ytan=xtan*yres/xres;
	m[0]=1.0/xtan; m[5]=1.0/ytan;
	m[11]=-1; // This in the fourth row is because Z is out of the screen in OpenGL, thus -z is the denominator for proper perspective
	m[10]=(zn+zf)/(zn-zf); // Third row produces numerator for depth buffer ordinate d=(Az+B)/-z, d(-zn)=1, d(-zf)=-1
	m[14]=2.0*zn*zf/(zn-zf);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

void set43fovzz(const GLfloat degs,const GLfloat zn,const GLfloat zf)
{ // Guarantees the 4:3 field will be visible on all monitor shapes
	glMatrixMode(GL_PROJECTION);
	GLfloat m[16]; // GL stores matrices column-first
	for (int i=0;i<16;i++) m[i]=0;
	GLfloat xtan=tan(degs degrees*0.5),
		ytan43=xtan*0.75,
		ytan=xtan*yres/xres;
	if (ytan<ytan43) {ytan=ytan43; xtan=ytan*xres/yres;}
	m[0]=1.0/xtan; m[5]=1.0/ytan;
	m[11]=-1; // This in the fourth row is because Z is out of the screen in OpenGL, thus -z is the denominator for proper perspective
	m[10]=(zn+zf)/(zn-zf); // Third row produces numerator for depth buffer ordinate d=(Az+B)/-z, d(-zn)=1, d(-zf)=-1
	m[14]=2.0*zn*zf/(zn-zf);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

#define setfov(A) setfovzz(A,0.1f,1000.0f)
#define set43fov(A) set43fovzz(A,0.1f,1000.0f)

void ReSizeGLScene(const int xs,int ys)
{
	if (ys<1) ys=1; // Prevent divide by zero later on...
	xres=xs; yres=ys;
	glViewport(0,0,xs,ys);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (gl_ortho) glOrtho(0,xs,ys,0,ys/2,-ys/2);
	else set43fov(90);
	glMatrixMode(GL_MODELVIEW);
}

void InitGL(const int xs,const int ys)	// This Will Be Called Right After The GL Window Is Created
{
	glClearColor(0.0f,0.0f,0.0f,0.0f);		// This Will Clear The Background Color To Black
	glClearDepth(1.0);							// Enables Clearing Of The Depth Buffer
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
	glDepthFunc(GL_LEQUAL);						// The Type Of Depth Test To Do
	glEnable(GL_DEPTH_TEST);					// Enables Depth Testing
	glShadeModel(GL_SMOOTH);					// Enables Smooth Color Shading (linear interpolation across polys and lines)
	ReSizeGLScene(xs,ys);
	glLoadIdentity(); // Set the GL_MODELVIEW matrix
}

float glversion(void)
{
	return atof((char *)glGetString(GL_VERSION));
}

void ccl(const unsigned c=0)
{ // Use glClearColor glClearDepth to set how this clearing works
	glFinish(); // Makes vsync work better
	#ifdef linux
	SDL_GL_SwapBuffers();
	#else
	SwapBuffers(hDC);
	#endif
	static unsigned oc=0x1000000;
	if (c!=oc) {glClearColor((float)(c>>16)/255.0,(float)((c>>8)&0xFF)/255.0,(float)(c&0xFF)/255.0,0); oc=c;}
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
	glLoadIdentity(); // Reset The View
}

void update(void)
{
	glFinish();
	#ifdef linux
	SDL_GL_SwapBuffers();
	#else
	SwapBuffers(hDC);
	#endif
	glLoadIdentity(); // Reset The View
}

void clearrect(const int ox,const int oy,const int xs,const int ys)
{ // Clears both the colour and depth buffers in a rectangular region
	glEnable(GL_SCISSOR_TEST); glScissor(ox,oy,xs,ys);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);
}

void cl(void)
{
	ccl(); ccl();
}

#ifndef linux
void plotdefaultmouse(const float b=1)
{ // Only works in ortho view
	GLboolean oldlight=glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(b,b,b);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+xres/200,my+0.1);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+0.1,my+xres/200);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+xres/100,my+xres/100);
	glEnd();
	if (oldlight) glEnable(GL_LIGHTING);
}

int contrastshade(const unsigned c)
{ // 0 for black, 1 for white
	if ((c&0xFF)*3+((c>>8)&0xFF)*6+((c>>16)&0xFF)*2>=0.5*0xFF*11) return 0;
	else return 1;
}

void plotvisiblemouse(void)
{ // Only works in ortho view
	unsigned c;
	glReadPixels(min(xres-1,mx+xres/200),yres-1-min(my+yres/200,yres-1),1,1,GL_RGB,GL_UNSIGNED_BYTE,&c);
	plotdefaultmouse(contrastshade(c));
}
#endif

void screenshotrect(const int ox,const int oy,const int xs,const int ys,const char *filename,const int png=1)
{ // filename has the ".bmp" extension in all these functions
	int n,x,y;
	FILE *out=fopen(filename,"wb");
	unsigned char seg1[18]={0x42,0x4D,0x46,0x0E,3,0,0,0,0,0,0x36,0,0,0,0x28,0,0,0},
		seg2[28]={1,0,0x18,0,0,0,0,0,0x10,0x0E,3,0,0x12,0x0B,0,0,0x12,0x0B,0,0,0,0,0,0,0,0,0,0};
	fwrite(seg1,1,18,out);
	fwrite(&xs,sizeof(int),1,out);
	fwrite(&ys,sizeof(int),1,out);
	fwrite(seg2,1,28,out);
	int pitch=(xs*3+3)/4*4;
	unsigned char *d=(unsigned char *)malloc(ys*pitch);
	glReadPixels(ox,yres-(oy+ys),xs,ys,GL_BGR,GL_UNSIGNED_BYTE,d);
	fwrite(d,ys*pitch,1,out);
	fclose(out); free(d);
	if (png) {char cmd[100]; sprintf(cmd,"bmp2png -9 -E %s",filename); system(cmd);}
}

void screenshot(const char *file,const int png=1) {screenshotrect(0,0,xres,yres,file,png);}

#include <rnd/eithertime.c>
void screenshotauto(void)
{
	char file[50];
	sprintf(file,"screenshot%05d.bmp",((int)(eithertime()/0.3))%100000);
	screenshot(file);
}

void paldata666(float *r,float *g,float *b)
{ // Only relevant for backwards compatibility with old 256-colour applications!
	int n;
	for (n=0;n<16;n++)
	{
		r[n]=((n&4)>0)?0.5+0.5*((n&8)>0):0;
		g[n]=((n&2)>0)?0.5+0.5*((n&8)>0):0;
		b[n]=((n&1)>0)?0.5+0.5*((n&8)>0):0;
	}
	r[8]=g[8]=b[8]=0.25;
	for (n=16;n<40;n++)
		r[n]=g[n]=b[n]=(float)(n-16)/23;
	for (n=40;n<256;n++)
	{
		r[n]=0.2*((n-40)%6);
		g[n]=0.2*((n-40)/6%6);
		b[n]=0.2*((n-40)/36);
	}
}

void setcolour(const int c)
{ // Emulates my 8-bit palette
	if (c<16)
	{
		if (c==8) glColor3f(0.25,0.25,0.25);
		else {float b=0.5+0.5*(c>8); glColor3f((c&4)?b:0,(c&2)?b:0,(c&1)?b:0);}
	}
	else if (c<40) {float b=(float)(c-16)/23; glColor3f(b,b,b);}
	else glColor3f(0.2*((c-40)%6),0.2*((c-40)/6%6),0.2*((c-40)/36));
}

unsigned *palto24bit(void)
{
	float r[256],g[256],b[256];
	paldata666(r,g,b);
	int n; unsigned *ret=(unsigned *)malloc(256*sizeof(unsigned));
	for (n=255;n>=0;n--) ret[n]=(((int)floor(r[n]*255.9))<<16)|(((int)floor(g[n]*255.9))<<8)|((int)floor(b[n]*255.9));
	return ret;
}

void initsmoothing(void)
{
	glEnable(GL_BLEND); // These four lines enable line smoothing, only glHint is optional
	glBlendFunc(GL_SRC_ALPHA_SATURATE,GL_ONE);
	glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
	glEnable(GL_POINT_SMOOTH);
}

void ortho_on(void)
{
	glPushMatrix(); // Save current model matrix state
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); // Save current projection matrix state
	glLoadIdentity();
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	glOrtho(v[0],v[0]+v[2],yres-v[1],yres-(v[1]+v[3]),yres/2,-yres/2);
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_DEPTH_TEST);
}

void ortho_off(void)
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
}

double ortho_pixelsize(void)
{ // Returns 1 screen pixel's logical size (erring on the average side) when in an orthogonal projection
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	GLdouble pm[16],vm[16],m[16];
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	glGetDoublev(GL_MODELVIEW_MATRIX,vm);
	int i,j,k; for (i=0;i<4;i++) for (j=0;j<4;j++)
	{
		m[i+j*4]=0;
		for (k=0;k<4;k++) m[i+j*4]+=pm[i+k*4]*vm[k+j*4];
	}
	for (i=0;i<4;i++)
	{
		m[i+0*4]*=0.5*v[2];
		m[i+1*4]*=0.5*v[3];
	}
	return 1.0/sqrt(fabs(m[0]*m[5]-m[1]*m[4])); // sqrt(det(screen->logical))
}

double radian_pixelsize(void)
{ // Radians (in X) per pixel at centre of screen in 3D mode
	GLdouble m[16];
	glGetDoublev(GL_PROJECTION_MATRIX,m);
	return 2.0/(m[0]*xres);
}

void onscreenpos(const double x,const double y,const double z,double *px,double *py)
{ // Translates 3D to my standard 2D coordinates (as used in bfwrite, ortho_on mode etc.)
	int i,v[4]; glGetIntegerv(GL_VIEWPORT,v);
	GLdouble pm[16],vm[16],a[4],b[4];
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	glGetDoublev(GL_MODELVIEW_MATRIX,vm);
	a[0]=x; a[1]=y; a[2]=z; a[3]=1;
	for (i=0;i<4;i++) b[i]=vm[i]*a[0]+vm[i+4]*a[1]+vm[i+8]*a[2]+vm[i+12]*a[3];
	for (i=0;i<2;i++) a[i]=pm[i]*b[0]+pm[i+4]*b[1]+pm[i+8]*b[2]+pm[i+12]*b[3];
	*px=v[0]+(a[0]*0.5+0.5)*v[2];
	*py=v[1]+(a[1]*-0.5+0.5)*v[3];
}

// 2D graphics functions
void putpix(const double x,const double y,const unsigned c)
{ // I pity the fool who uses this in a loop
	glBegin(GL_POINTS);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
	glEnd();
}

void putpix_inner(const double x,const double y,const unsigned c)
{ // glBegin(GL_POINTS)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
}

void line(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // Not really recommended, more for 2D compatibility
	GLboolean oldlight=glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y1);
	glEnd();
	if (oldlight) glEnable(GL_LIGHTING);
}

void line_inner(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // glBegin(GL_LINES)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y1);
}

void rect(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // Probably slow, use rect_inner if possible
	glBegin(GL_QUADS);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
	glEnd();
}

void rect_inner(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // glBegin(GL_QUADS)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
}

void box(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // Probably slow, use box_inner (which doesn't exist yet) if possible
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
	glEnd();
}

void boxa(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // Probably slow, use boxa_inner (which doesn't exist yet) if possible
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y0); glVertex2d(x1,y1); glVertex2d(x0,y1);
	glEnd();
}

void circlefill(const double x,const double y,const double r,const unsigned c)
{ // No need to do "_inner" of this, need to start new TRIANGLE_FAN anyway
	int n,dot_sides=(int)(5.0*sqrt(r/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=M_TWOPI/dot_sides,th;
	glBegin(GL_TRIANGLE_FAN);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
	for (n=dot_sides,th=0;n>=0;n--,th+=step)
		glVertex2d(x+r*sin(th),y+r*cos(th));
	glEnd();
}

void circle(const double x,const double y,const double r,const unsigned c)
{ // Neither of these are particularly fast right now
	int n,dot_sides=(int)(5.0*sqrt(r/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=M_TWOPI/dot_sides,th;
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	for (n=dot_sides,th=0;n>=0;n--,th+=step)
		glVertex2d(x+r*sin(th),y+r*cos(th));
	glEnd();
}

void circular(const double x,const double y,const double r1,const double r2,const double th1,const double th2,const unsigned c,const unsigned char a=0xFF)
{ // theta should be as on a clock for inverted Y-coordinate systems
	int n,dot_sides=(int)max(1,((th2-th1)/M_TWOPI)*5.0*sqrt(r2/ortho_pixelsize()));
	double step=(th2-th1)/dot_sides,th;
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,a);
	if (r1==0)
	{
		glBegin(GL_TRIANGLE_FAN);
		glVertex2d(x,y);
		for (n=dot_sides,th=th1;n>=0;n--,th+=step)
			glVertex2d(x+r2*sin(th),y-r2*cos(th));
		glEnd();
	}
	else
	{
		float sth,cth;
		glBegin(GL_QUAD_STRIP);
		for (n=dot_sides,th=th1;n>=0;n--,th+=step)
		{
			glVertex2d(x+r1*(sth=sin(th)),y-r1*(cth=cos(th)));
			glVertex2d(x+r2*sth,y-r2*cth);
		}
		glEnd();
	}
}

#ifdef SBCLIB_BMP
#include <rnd/sjbBuild2DMipmaps.c>
#include <rnd/enclosingpoweroftwo.c>

GLuint bmptexture(const char *filename)
{ // Texture from 24-bit BMP
	GLuint ret;
	glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_2D,ret); // Selects the current texture
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	bmp b=bmpload(filename);
	if (b.xs==0) {b.xs=b.ys=1; b.d=calloc(3,1); b.d[0]=0x80;}
	int tx=enclosingpoweroftwo(b.xs),ty=enclosingpoweroftwo(b.ys);
	int pitch=(tx*3+3)/4*4; // Round up to 4-byte boundary
	unsigned char *d=calloc(ty*pitch,1),*br,*dr;
	int x,y;
	for (y=b.ys-1;y>=0;y--)
	{
		dr=d+(b.ys-1-y)*pitch; br=b.d+y*b.xs*3;
		for (x=b.xs-1;x>=0;x--) {dr[x*3]=br[x*3+2]; dr[x*3+1]=br[x*3+1]; dr[x*3+2]=br[x*3];}
	}
	sjbBuild2DMipmaps(3,tx,ty,d);
	bmpfree(b); free(d);
	return ret;
}
#endif

void aniso(void)
{
	static int supported=-1;
	if (supported<0) supported=glIsSupported("GL_EXT_texture_filter_anisotropic");
	if (!supported) return;
	static float maxaniso=0;
	//if (KEY(VK_A)) maxaniso=1; else maxaniso=0;
	if (!maxaniso)
	{
		glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT,&maxaniso);
		//FILE *out=fopen("aniso.log","at"); fprintf(out,"maxaniso = %lg\n",maxaniso); fclose(out);
	}
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,maxaniso);
}

#ifndef linux
void vsync(const int i)
{ // NB: the Linux equivalent of this is the GLX_SGI_swap_control extension I think
	static int supported=-1;
	if (supported<0) supported=glIsSupported("WGL_EXT_swap_control");
	if (!supported) return;
	BOOL static (APIENTRY *wglSwapIntervalEXT)(int interval)=NULL; // APIENTRY resolves to __stdcall
	if (!wglSwapIntervalEXT) wglSwapIntervalEXT=(BOOL (APIENTRY *)(int))wglGetProcAddress("wglSwapIntervalEXT");
	wglSwapIntervalEXT(i);
}
#endif

// 3D graphics functions
void sphere(const double ox,const double oy,const double oz,const double r,const unsigned col,const int m=10)
{
	int n,i; double *c=(double *)malloc((m+1)*sizeof(double)),*s=(double *)malloc((m+1)*sizeof(double)),th;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glBegin(GL_QUADS);
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m-1;n>=0;n--) for (i=m-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i],s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy+r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy+r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy+r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy+r*s[n]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i],-s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy-r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],-s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy-r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],-s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy-r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],-s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy-r*s[n]*s[i+1],oz+r*c[i+1]);
	}
	glEnd();
	free(c); free(s);
}

void cylinder_inner(const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double r,const unsigned col,const int m)
{ // Put this inside glBegin(GL_QUADS), m is divisions in 180 degrees
	double ax=x2-x1,ay=y2-y1,az=z2-z1,bx,by,bz,cx,cy,cz,k,dx,dy,dz;
	k=ay*ay+az*az;
	if (k>1e-12)
	{ // b := a x (1,0,0)
		bx=0; by=az; bz=-ay;
	}
	else
	{ // b := a x (0,1,0)
		bx=-az; by=0; bz=ax;
		k=ax*ax+az*az;
	}
	k=1.0/sqrt(k); bx*=k; by*=k; bz*=k;
	// c := a x b
	cx=ay*bz-az*by; cy=az*bx-ax*bz; cz=ax*by-ay*bx;
	k=1.0/sqrt(cx*cx+cy*cy+cz*cz); cx*=k; cy*=k; cz*=k;
	int n; double *c=(double *)malloc((m+1)*sizeof(double)),*s=(double *)malloc((m+1)*sizeof(double)),th;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m-1;n>=0;n--)
	{
		glNormal3d(dx=c[n]*bx+s[n]*cx,dy=c[n]*by+s[n]*cy,dz=c[n]*bz+s[n]*cz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx=c[n+1]*bx+s[n+1]*cx,dy=c[n+1]*by+s[n+1]*cy,dz=c[n+1]*bz+s[n+1]*cz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx=-c[n]*bx-s[n]*cx,dy=-c[n]*by-s[n]*cy,dz=-c[n]*bz-s[n]*cz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx=-c[n+1]*bx-s[n+1]*cx,dy=-c[n+1]*by-s[n+1]*cy,dz=-c[n+1]*bz-s[n+1]*cz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
	}
	free(c); free(s);
}

void cube_inner(const double x,const double y,const double z,const double s)
{ // Must be within glBegin(GL_QUADS)...
	glNormal3f(0,0,-1);
	glVertex3d(x,y,z); glVertex3d(x+s,y,z); glVertex3d(x+s,y+s,z); glVertex3d(x,y+s,z);
	glNormal3f(0,0,1);
	glVertex3d(x,y,z+s); glVertex3d(x+s,y,z+s); glVertex3d(x+s,y+s,z+s); glVertex3d(x,y+s,z+s);
	glNormal3f(0,-1,0);
	glVertex3d(x,y,z); glVertex3d(x+s,y,z); glVertex3d(x+s,y,z+s); glVertex3d(x,y,z+s);
	glNormal3f(0,1,0);
	glVertex3d(x,y+s,z); glVertex3d(x+s,y+s,z); glVertex3d(x+s,y+s,z+s); glVertex3d(x,y+s,z+s);
	glNormal3f(-1,0,0);
	glVertex3d(x,y,z); glVertex3d(x,y+s,z); glVertex3d(x,y+s,z+s); glVertex3d(x,y,z+s);
	glNormal3f(1,0,0);
	glVertex3d(x+s,y,z); glVertex3d(x+s,y+s,z); glVertex3d(x+s,y+s,z+s); glVertex3d(x+s,y,z+s);
}

void cubetex_inner(const double x,const double y,const double z,const double s)
{ // Also needs 2D texturing (one texture per cube face)
	glNormal3f(0,0,-1);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z); glTexCoord2f(0,1); glVertex3d(x,y+s,z);
	glNormal3f(0,0,1);
	glTexCoord2f(0,0); glVertex3d(x,y,z+s); glTexCoord2f(1,0); glVertex3d(x+s,y,z+s); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y+s,z+s);
	glNormal3f(0,-1,0);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y,z); glTexCoord2f(1,1); glVertex3d(x+s,y,z+s); glTexCoord2f(0,1); glVertex3d(x,y,z+s);
	glNormal3f(0,1,0);
	glTexCoord2f(0,0); glVertex3d(x,y+s,z); glTexCoord2f(1,0); glVertex3d(x+s,y+s,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y+s,z+s);
	glNormal3f(-1,0,0);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x,y+s,z); glTexCoord2f(1,1); glVertex3d(x,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y,z+s);
	glNormal3f(1,0,0);
	glTexCoord2f(0,0); glVertex3d(x+s,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y+s,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x+s,y,z+s);
}

void cuboid_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1)
{ // Must be within glBegin(GL_QUADS)...
	glNormal3f(0,0,-1);
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0); glVertex3d(x0,y1,z0);
	glNormal3f(0,0,1);
	glVertex3d(x0,y0,z1); glVertex3d(x1,y0,z1); glVertex3d(x1,y1,z1); glVertex3d(x0,y1,z1);
	glNormal3f(0,-1,0);
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0); glVertex3d(x1,y0,z1); glVertex3d(x0,y0,z1);
	glNormal3f(0,1,0);
	glVertex3d(x0,y1,z0); glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1); glVertex3d(x0,y1,z1);
	glNormal3f(-1,0,0);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y1,z0); glVertex3d(x0,y1,z1); glVertex3d(x0,y0,z1);
	glNormal3f(1,0,0);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1); glVertex3d(x1,y0,z1);
}

void cuboidtex_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1,const double tsc)
{ // Also needs 2D texturing (texture repeats at tsc in size)
	double tx=(x1-x0)/tsc,ty=(y1-y0)/tsc,tz=(z1-z0)/tsc;
	glNormal3f(0,0,-1);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(tx,0); glVertex3d(x1,y0,z0); glTexCoord2d(tx,ty); glVertex3d(x1,y1,z0); glTexCoord2d(0,ty); glVertex3d(x0,y1,z0);
	glNormal3f(0,0,1);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z1); glTexCoord2d(tx,0); glVertex3d(x1,y0,z1); glTexCoord2d(tx,ty); glVertex3d(x1,y1,z1); glTexCoord2d(0,ty); glVertex3d(x0,y1,z1);
	glNormal3f(0,-1,0);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(tx,0); glVertex3d(x1,y0,z0); glTexCoord2d(tx,tz); glVertex3d(x1,y0,z1); glTexCoord2d(0,tz); glVertex3d(x0,y0,z1);
	glNormal3f(0,1,0);
	glTexCoord2d(0,0); glVertex3d(x0,y1,z0); glTexCoord2d(tx,0); glVertex3d(x1,y1,z0); glTexCoord2d(tx,tz); glVertex3d(x1,y1,z1); glTexCoord2d(0,tz); glVertex3d(x0,y1,z1);
	glNormal3f(-1,0,0);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(ty,0); glVertex3d(x0,y1,z0); glTexCoord2d(ty,tz); glVertex3d(x0,y1,z1); glTexCoord2d(0,tz); glVertex3d(x0,y0,z1);
	glNormal3f(1,0,0);
	glTexCoord2d(0,0); glVertex3d(x1,y0,z0); glTexCoord2d(ty,0); glVertex3d(x1,y1,z0); glTexCoord2d(ty,tz); glVertex3d(x1,y1,z1); glTexCoord2d(0,tz); glVertex3d(x1,y0,z1);
}

void cuboid_wireframe_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1)
{ // Must be within glBegin(GL_LINES)
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0);
	glVertex3d(x0,y0,z1); glVertex3d(x1,y0,z1);
	glVertex3d(x0,y1,z0); glVertex3d(x1,y1,z0);
	glVertex3d(x0,y1,z1); glVertex3d(x1,y1,z1);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y1,z0);
	glVertex3d(x0,y0,z1); glVertex3d(x0,y1,z1);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0);
	glVertex3d(x1,y0,z1); glVertex3d(x1,y1,z1);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y0,z1);
	glVertex3d(x0,y1,z0); glVertex3d(x0,y1,z1);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y0,z1);
	glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1);
}

void sphere_octant_inner(const double ox,const double oy,const double oz,const double r,
	const int kx,const int ky,const int kz,
	const unsigned col,const int m)
{ // m is number of divisions in 180 degrees (must be even)
	int n,i; double th,
		*c=(double *)malloc((m/2+1)*sizeof(double)),
		*s=(double *)malloc((m/2+1)*sizeof(double));
	for (n=m/2;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m/2-1;n>=0;n--) for (i=m/2-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i]*kx,s[n]*s[i]*ky,c[i]*kz);
		glVertex3d(ox+r*c[n]*s[i]*kx,oy+r*s[n]*s[i]*ky,oz+r*c[i]*kz);
		glNormal3d(c[n+1]*s[i]*kx,s[n+1]*s[i]*ky,c[i]*kz);
		glVertex3d(ox+r*c[n+1]*s[i]*kx,oy+r*s[n+1]*s[i]*ky,oz+r*c[i]*kz);
		glNormal3d(c[n+1]*s[i+1]*kx,s[n+1]*s[i+1]*ky,c[i+1]*kz);
		glVertex3d(ox+r*c[n+1]*s[i+1]*kx,oy+r*s[n+1]*s[i+1]*ky,oz+r*c[i+1]*kz);
		glNormal3d(c[n]*s[i+1]*kx,s[n]*s[i+1]*ky,c[i+1]*kz);
		glVertex3d(ox+r*c[n]*s[i+1]*kx,oy+r*s[n]*s[i+1]*ky,oz+r*c[i+1]*kz);
	}
	free(c); free(s);
}

void sphere_inner(const double ox,const double oy,const double oz,const double r,const unsigned col,const int m)
{ // m is number of divisions in 180 degrees
	int n,i; double th,
		*c=(double *)malloc((m+1)*sizeof(double)),
		*s=(double *)malloc((m+1)*sizeof(double));
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m-1;n>=0;n--) for (i=m-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i],s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy+r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy+r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy+r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy+r*s[n]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i],-s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy-r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],-s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy-r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],-s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy-r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],-s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy-r*s[n]*s[i+1],oz+r*c[i+1]);
	}
	free(c); free(s);
}

// Advanced OpenGL graphics
void blur(float r)
{
	static GLuint scr=0; if (!scr) glGenTextures(1,&scr);
	GLboolean oldtex2d=glIsEnabled(GL_TEXTURE_2D),oldlight=glIsEnabled(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D); glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D,scr);
	int xs,ys; for (xs=1;xs<xres;xs<<=1); for (ys=1;ys<yres;ys<<=1);
	GLfloat tx=(GLfloat)xres/xs,ty=(GLfloat)yres/ys,tr;
	ortho_on();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	for (;r>=1;r*=0.5)
	{
		glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,0,0,xs,ys,0); // Not GL_RGBA8, that causes weird mallards
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glBegin(GL_QUADS);
		glColor4f(1,1,1,1.0/3);
		tr=(GLfloat)r/xs;
		glTexCoord2f(tr,ty); glVertex2f(0,0);
		glTexCoord2f(tx+tr,ty); glVertex2f(xres,0);
		glTexCoord2f(tx+tr,0); glVertex2f(xres,yres);
		glTexCoord2f(tr,0); glVertex2f(0,yres);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(-tr,ty); glVertex2f(0,0);
		glTexCoord2f(tx-tr,ty); glVertex2f(xres,0);
		glTexCoord2f(tx-tr,0); glVertex2f(xres,yres);
		glTexCoord2f(-tr,0); glVertex2f(0,yres);
		/*glColor4f(1,1,1,1.0/3);
		glTexCoord2f(0,ty); glVertex2f(0,r);
		glTexCoord2f(tx,ty); glVertex2f(xres,r);
		glTexCoord2f(tx,0); glVertex2f(xres,yres+r);
		glTexCoord2f(0,0); glVertex2f(0,yres+r);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(0,ty); glVertex2f(r,r);
		glTexCoord2f(tx,ty); glVertex2f(xres+r,r);
		glTexCoord2f(tx,0); glVertex2f(xres+r,yres+r);
		glTexCoord2f(0,0); glVertex2f(r,yres+r);*/
		glEnd();
		glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,0,0,xs,ys,0); // Not GL_RGBA8, that causes weird mallards
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glBegin(GL_QUADS);
		glColor4f(1,1,1,1.0/3);
		tr=(GLfloat)r/ys;
		glTexCoord2f(0,ty+tr); glVertex2f(0,0);
		glTexCoord2f(tx,ty+tr); glVertex2f(xres,0);
		glTexCoord2f(tx,tr); glVertex2f(xres,yres);
		glTexCoord2f(0,tr); glVertex2f(0,yres);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(0,ty-tr); glVertex2f(0,0);
		glTexCoord2f(tx,ty-tr); glVertex2f(xres,0);
		glTexCoord2f(tx,-tr); glVertex2f(xres,yres);
		glTexCoord2f(0,-tr); glVertex2f(0,yres);
		glEnd();
	}
	//glDeleteTextures(1,&scr); // Creating and deleting the texture object was the slow part!!
	ortho_off();
	glBlendFunc(GL_ONE,GL_ZERO);
	if (!oldtex2d) glDisable(GL_TEXTURE_2D);
	if (oldlight) glEnable(GL_LIGHTING);
}

#include <rnd/sjbBuild2DMipmaps.c>

GLuint makecubemap(void colfunc(double,double,double,unsigned char *),const unsigned res,const unsigned aa=1)
{ // colfunc accepts "direction cosines" and writes the RGB bytes starting at the supplied pointer, res should be power of two
	int x,y,i; GLuint ret; glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_CUBE_MAP,ret);
	unsigned char *datapz=(unsigned char *)malloc(res*res*3),*datanz=(unsigned char *)malloc(res*res*3),
		*datapy=(unsigned char *)malloc(res*res*3),*datany=(unsigned char *)malloc(res*res*3),
		*datapx=(unsigned char *)malloc(res*res*3),*datanx=(unsigned char *)malloc(res*res*3);
	double fx,fy,k,rh=0.5*res;
	if (aa<=1) for (y=res-1;y>=0;y--) for (x=res-1;x>=0;x--)
	{
		fx=(x+0.5-rh)/rh; fy=(y+0.5-rh)/rh; k=1.0/sqrt(1.0+fx*fx+fy*fy);
		fx*=k; fy*=k; i=(y*res+x)*3;
		colfunc(fx,-fy,k,datapz+i);
		colfunc(-fx,-fy,-k,datanz+i);
		colfunc(fx,k,fy,datapy+i);
		colfunc(fx,-k,-fy,datany+i);
		colfunc(k,-fy,-fx,datapx+i);
		colfunc(-k,-fy,fx,datanx+i);
	}
	else // Antialiasing, pretty CPU-intensive!
	{
		int n,aas=aa*aa; unsigned char s[18],c; double ss[18],iaas=1.0/aas;
		for (y=res-1;y>=0;y--) for (x=res-1;x>=0;x--)
		{
			for (n=17;n>=0;n--) ss[n]=0;
			for (n=aas-1;n>=0;n--)
			{
				i=aa-1-n/aa+n%aa*aa;
				fx=(x-rh+(0.5+n)/aas)/rh; fy=(y-rh+(0.5+i)/aas)/rh;
				k=1.0/sqrt(1.0+fx*fx+fy*fy); fx*=k; fy*=k;
				colfunc(fx,-fy,k,s);
				colfunc(-fx,-fy,-k,s+3);
				colfunc(fx,k,fy,s+6);
				colfunc(fx,-k,-fy,s+9);
				colfunc(k,-fy,-fx,s+12);
				colfunc(-k,-fy,fx,s+15);
				for (i=17;i>=0;i--) ss[i]+=s[i];
			}
			i=(y*res+x)*3;
			for (n=2;n>=0;n--)
			{
				datapz[i+n]=c=(unsigned char)(iaas*ss[n]);
				datanz[i+n]=c=(unsigned char)(iaas*ss[n+3]);
				datapy[i+n]=c=(unsigned char)(iaas*ss[n+6]);
				datany[i+n]=c=(unsigned char)(iaas*ss[n+9]);
				datapx[i+n]=c=(unsigned char)(iaas*ss[n+12]);
				datanx[i+n]=c=(unsigned char)(iaas*ss[n+15]);
			}
		}
	}
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapz);
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datanz);
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapy);
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datany);
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapx);
	//glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datanx);
	sjbBuildCubeMipmaps(3,res,datapx,datanx,datapy,datany,datapz,datanz);
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR); // GL_LINEAR for glTexImage2D
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	free(datapz); free(datanz); free(datapy); free(datany); free(datapx); free(datanx);
	return ret;
}

// Need another function that generates a cube map from a function that plots a scene

void skysphere(const GLuint cubemap)
{ // Sky sphere; put the inverse of the viewer rotation matrix in the GL_TEXTURE matrix first! (similar for cubemaps/reflection maps)
	glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);
	glPushMatrix(); glLoadIdentity();
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP,cubemap);
	glBegin(GL_QUADS);
	glColor3f(1,1,1);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glEnd();
	glClear(GL_DEPTH_BUFFER_BIT);
	glPopMatrix();
	glPopAttrib();
}

#define SBCLIB_GRAPHICS
#define SBCLIB_GRAPHICS_OPENGL
#define INCLUDED_SBCLIB_GRAPHICS_OPENGL
#endif
